//
// This file is part of AlphaCam II.
//
// AlphaCam II is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AlphaCam II is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AlphaCam II.  If not, see <http://www.gnu.org/licenses/>.
//
//
//  NLGLCamera.m
//  AlphaCam II
//
//  Created by Nicholas Levin on 5/6/11.
//  Copyright 2011 Studio Andromeda. All rights reserved.
//

#import "NLGLCamera.h"

#import <OpenGLES/EAGL.h>
#import <OpenGLES/ES2/gl.h>
#import <OpenGLES/ES2/glext.h>

@implementation NLGLCamera

#pragma mark -
#pragma mark Initialization and teardown

- (id)init; 
{
	if (!(self = [super init]))
		return nil;
	
	// Grab the back-facing camera
	AVCaptureDevice *backFacingCamera = nil;
	NSArray *devices = [AVCaptureDevice devicesWithMediaType:AVMediaTypeVideo];
    
	for (AVCaptureDevice *device in devices) 
	{
		if ([device position] == AVCaptureDevicePositionBack) 
		{
			backFacingCamera = device;
		}
	}
	
	// Create the capture session
	captureSession = [[AVCaptureSession alloc] init];
	
	// Add the video input	
	NSError *error = nil;
	videoInput = [[AVCaptureDeviceInput alloc] initWithDevice:backFacingCamera error:&error];
	if ([captureSession canAddInput:videoInput]) 
	{
		[captureSession addInput:videoInput];
	}
    
	// allocate and initialize the AVCaptureVideoPreviewLayer
	//[self videoPreviewLayer];
    
	// Add the video frame output	
	videoOutput = [[AVCaptureVideoDataOutput alloc] init];
	
    //[videoOutput setAlwaysDiscardsLateVideoFrames:YES];      // Might want to set this to NO when we're recording
    // WAS setMinFrameDuration: in iOS 4.1+, until iOS 5.0. Make sure this calls that instead of AVCaptureConnection's setVideoMinFrameDuration: when this app runs on iOS 4.1+ !
    [videoOutput setMinFrameDuration:CMTimeMake(1,30)];
                                                            //      24 fps is akin to movies
                                                            //      30 fps is akin to most video games
                                                            //      60 fps is the most the eye can percieve
                                                            //      less than 24 looks jittery
    
    //[videoOutput setAlwaysDiscardsLateVideoFrames:NO];
    //[videoOutput setVideoMinFrameDuration:CMTimeMake(30,1)];   // if 24 fps is too inconsistent, try this
    
	// Use RGB frames instead of YUV to ease color processing
	[videoOutput setVideoSettings:[NSDictionary dictionaryWithObject:[NSNumber numberWithInt:kCVPixelFormatType_32BGRA] forKey:(id)kCVPixelBufferPixelFormatTypeKey]];
    [videoOutput setSampleBufferDelegate:self queue:dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0)];
    
    // In the WWDC examples, we create our own background queue for video and leave the main queue available for audio capture.
    // However, the main queue is necessary for OpenGL to capture camera output. Otherwise, we get undefined greyness.
    
	if ([captureSession canAddOutput:videoOutput])
	{
		[captureSession addOutput:videoOutput];
	}
	else
	{
		printf("Couldn't add video output\n");
	}

	// Start capturing
	[captureSession setSessionPreset:AVCaptureSessionPreset640x480];
	if (![captureSession isRunning])
	{
        [captureSession startRunning];
	}
	
	return self;
}

- (void)dealloc 
{
	[captureSession stopRunning];

}

#pragma mark -
#pragma mark OpenGL Screenshot Methods


- (void)imageSavedToPhotosAlbum:(UIImage *)image didFinishSavingWithError:(NSError *)error contextInfo:(void *)contextInfo
{  
    NSString *message;  
    NSString *title;  
    
    if (!error) {  
        title = NSLocalizedString(@"SaveSuccessTitle", @"");  
        message = NSLocalizedString(@"SaveSuccessMessage", @"");  
    } else {  
        title = NSLocalizedString(@"SaveFailedTitle", @"");  
        message = [error description];  
    }
    UIAlertView *alert = [[UIAlertView alloc] initWithTitle:title  
                                                    message:message  
                                                   delegate:nil  
                                          cancelButtonTitle:NSLocalizedString(@"ButtonOK", @"")  
                                          otherButtonTitles:nil];  
    [alert show];  
}

-(void)saveImageToLocalPhotoAlbum:(UIImage*)resultUIImage
{    
    // hacky workaround to preserve quality before saving to the Photos Album
    NSData* imdata = UIImagePNGRepresentation ( resultUIImage );
    resultUIImage = [UIImage imageWithData:imdata];
    UIImageWriteToSavedPhotosAlbum(resultUIImage, self, @selector(imageSavedToPhotosAlbum: didFinishSavingWithError: contextInfo:), nil);
}

-(void)openGLViewScreenshot:(UIView*)eaglview
{	
	
    // Get the size of the backing CAEAGLLayer
    GLint backingWidth, backingHeight;
    //glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer); // only have one, so don't worry about it
    glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_WIDTH, &backingWidth);
    glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_HEIGHT, &backingHeight);
    NSInteger x = 0, y = 0, width = backingWidth, height = backingHeight;
    NSInteger dataLength = width * height * 4;
    // makes more sense to preallocate these calculations in real world usage
    
    GLubyte *data = malloc(dataLength * sizeof(GLubyte));     // our screenshot buffer
	
    // Read pixel data from the framebuffer
    glPixelStorei(GL_PACK_ALIGNMENT, 4);                                // forces 4 byte alignment - do we need this?
    glReadPixels(x, y, width, height, GL_RGBA, GL_UNSIGNED_BYTE, data); // this is the most important bit!
    
    // now we perform a series of C-based Core Graphics calls to change the raw pixel buffer into a bitmap... (assume a max of 31 fps on current devices)
	
    // Create a CGImage with the pixel data
    // Since our OpenGL ES content is opaque, we use kCGImageAlphaNoneSkipLast to ignore the alpha channel. Otherwise, use kCGImageAlphaPremultipliedLast
    CGDataProviderRef ref = CGDataProviderCreateWithData(NULL, data, dataLength, NULL);
    
    // instead of the following two steps, we could skip straight to a CVPixelBufferCreateWithBytes operation without the middleman of producing UIImages...
    
    CGColorSpaceRef colorspace = CGColorSpaceCreateDeviceRGB();
    CGImageRef iref = CGImageCreate(width, height, 8, 32, width * 4, colorspace, kCGBitmapByteOrder32Big | kCGImageAlphaNoneSkipLast,
                                    ref, NULL, true, kCGRenderingIntentDefault);
	
    // OpenGL ES measures data in PIXELS
    // Create a graphics context with the target size measured in POINTS
    NSInteger widthInPoints, heightInPoints;
    if (NULL != UIGraphicsBeginImageContextWithOptions) {
        // On iOS 4 and later, use UIGraphicsBeginImageContextWithOptions to take the scale into consideration
        // Set the scale parameter to your OpenGL ES view's contentScaleFactor
        // so that you get a high-resolution snapshot when its value is greater than 1.0
        
        // this takes an objective-c messaging call! we should retain this in a more local scope with our precalculations
        CGFloat scale = eaglview.contentScaleFactor;
        
        // precalculate the following two variables
        widthInPoints = width / scale;
        heightInPoints = height / scale;
        
        UIGraphicsBeginImageContextWithOptions(CGSizeMake(widthInPoints, heightInPoints), NO, scale);
    }
    else {
        // On iOS prior to 4, fall back to use UIGraphicsBeginImageContext
        
        // precalculate the following two variables
        widthInPoints = width;
        heightInPoints = height;
        
        UIGraphicsBeginImageContext(CGSizeMake(widthInPoints, heightInPoints));
    }
	
    CGContextRef cgcontext = UIGraphicsGetCurrentContext();
	
    // UIKit coordinate system is upside down to GL/Quartz coordinate system
    // Flip the CGImage by rendering it to the flipped bitmap context
    // The size of the destination area is measured in POINTS
    CGContextSetBlendMode(cgcontext, kCGBlendModeCopy);
    CGContextDrawImage(cgcontext, CGRectMake(0.0, 0.0, widthInPoints, heightInPoints), iref);
	
    // Retrieve the UIImage from the current context
    UIImage *image = UIGraphicsGetImageFromCurrentImageContext();
	
    UIGraphicsEndImageContext();
	
    // Clean up
    free(data);
    CFRelease(ref);
    CFRelease(colorspace);
    CGImageRelease(iref);
	
	// Save the image to our user's photo album
    [self saveImageToLocalPhotoAlbum:image];
}


#pragma mark -
#pragma mark AVCaptureVideoDataOutputSampleBufferDelegate

// the one member of this delegate - this is how we output frames to OpenGL
- (void)captureOutput:(AVCaptureOutput *)captureOutput didOutputSampleBuffer:(CMSampleBufferRef)sampleBuffer fromConnection:(AVCaptureConnection *)connection
{    
    CFRetain(sampleBuffer);
    
    // necessary to keep the video processing on a separate thread; otherwise, we could do without the CFRunLoop call.
    CFRunLoopPerformBlock(CFRunLoopGetMain(), kCFRunLoopCommonModes, ^(void){

        [self.delegate processNewCameraFrame:sampleBuffer];

        //[self.delegate drawFrame];
    });
}

#pragma mark -
#pragma mark Accessors

@synthesize delegate;

@end
