//
// This file is part of AlphaCam II.
//
// AlphaCam II is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AlphaCam II is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AlphaCam II.  If not, see <http://www.gnu.org/licenses/>.
//
//
//  NLGLCameraViewController.m
//  AlphaCam II
//
//  Created by Nicholas Levin on 5/6/11.
//  Copyright 2011 Studio Andromeda. All rights reserved.
//

#import "NLGLCameraViewController.h"
#import "ESRenderer.h"
#import "NLUtility.h"


// NOTE: we can't use more than 64 uniform values in a single shader. Short of moving this entire matrix into a texture, this is as good as it gets for our Bayer Matrix-based dithering shader.
static const GLfloat NL_ditherShader_ditherMatrix[64] = {
    0.000000f, 0.500000f, 0.125000f, 0.625000f, 0.031250f, 0.531250f, 0.156250f, 0.656250f,   
    0.750000f, 0.250000f, 0.875000f, 0.375000f, 0.781250f, 0.281250f, 0.906250f, 0.406250f,  
    0.187500f, 0.687500f, 0.062500f, 0.562500f, 0.218750f, 0.718750f, 0.093750f, 0.593750f, 
    0.937500f, 0.437500f, 0.812500f, 0.312500f, 0.968750f, 0.468750f, 0.843750f, 0.343750f,
    0.046875f, 0.546875f, 0.171875f, 0.671875f, 0.015625f, 0.515625f, 0.140625f, 0.640625f,   
    0.796875f, 0.296875f, 0.921875f, 0.421875f, 0.765625f, 0.265625f, 0.890625f, 0.390625f,
    0.234375f, 0.734375f, 0.109375f, 0.609375f, 0.203125f, 0.703125f, 0.078125f, 0.578125f,
    0.984375f, 0.484375f, 0.859375f, 0.359375f, 0.953125f, 0.453125f, 0.828125f, 0.328125f 
};


// Uniform index.
enum {
    UNIFORM_VIDEOFRAME,
    UNIFORM_LUTFRAME,
    UNIFORM_DITHERMATRIX,
    NUM_UNIFORMS
};
GLint uniforms[NUM_UNIFORMS];

// Attribute index.
enum {
    ATTRIB_VERTEX,
    ATTRIB_TEXTUREPOSITON,
    NUM_ATTRIBUTES
};

@implementation NLGLCameraViewController


#pragma mark -
#pragma mark Initialization and teardown

- (id)initWithScreen:(UIScreen *)newScreenForDisplay;
{
    if ((self = [super initWithNibName:nil bundle:nil])) 
	{        
		screenForDisplay = newScreenForDisplay;
    }
    return self;
}

- (void)loadView 
{
	CGRect applicationFrame = [screenForDisplay applicationFrame];	
	CGRect mainScreenFrame = [[UIScreen mainScreen] applicationFrame];	
	UIView *primaryView = [[UIView alloc] initWithFrame:mainScreenFrame];
	self.view = primaryView;

	glView = [[NLGLView alloc] initWithFrame:CGRectMake(0.0f, 0.0f, applicationFrame.size.width, applicationFrame.size.height)];	
	[self.view addSubview:glView];
    glView.delegate = self;
	
    videoFrameTexture = [self createTextureUsingWidth:640 Height:480];
	
    [self loadVertexShader:@"DirectDisplayShader" fragmentShader:@"DirectDisplayShader" forProgram:&directDisplayProgram];
	[self loadVertexShader:@"DitherShader" fragmentShader:@"DitherShader" forProgram:&ditherProgram];
    
	// Set up the toolbar at the bottom of the screen
	UISegmentedControl *displayModeControl = [[UISegmentedControl alloc] initWithItems:[NSArray arrayWithObjects:NSLocalizedString(@"Video", nil), NSLocalizedString(@"Simple Dither", nil), nil]];
	displayModeControl.segmentedControlStyle = UISegmentedControlStyleBar;
	displayModeControl.selectedSegmentIndex = 0;
	[displayModeControl addTarget:self action:@selector(handleSwitchOfDisplayMode:) forControlEvents:UIControlEventValueChanged];
	
	UIBarButtonItem *item = [[UIBarButtonItem alloc] initWithCustomView:displayModeControl];
	displayModeControl.frame = CGRectMake(0.0f, 5.0f, 300.0f, 30.0f);
	
	NSArray *theToolbarItems = [NSArray arrayWithObjects:item, nil];
	
	UIToolbar *lowerToolbar = [[UIToolbar alloc] initWithFrame:CGRectMake(0.0f, self.view.frame.size.height - 44.0f, self.view.frame.size.width, 44.0f)];
	lowerToolbar.autoresizingMask = UIViewAutoresizingFlexibleWidth;
	lowerToolbar.tintColor = [UIColor blackColor];
	
	[lowerToolbar setItems:theToolbarItems];
	
	[self.view addSubview:lowerToolbar];
	
	camera = [[NLGLCamera alloc] init];
	camera.delegate = self;
	[self cameraHasConnected];
    
    // set up our video recorder object
    videoRecorder = [[NLGLCameraVideoRecorder alloc] init];
    
    
    
    // Load the LUT Texture into the shader
    [self loadLUTTexture];
    
    // Initialize OpenGL parameters for this context
    [self initGLConstantValues];
    
    
	
    pboIndex = 0;

    glViewport(0, 0, glView.backingWidth, glView.backingHeight);
    glBindTexture(GL_TEXTURE_2D, videoFrameTexture);
    
    displayModeHasChanged = YES;
}

- (void)didReceiveMemoryWarning 
{    
    [super didReceiveMemoryWarning];
}

- (void)dealloc 
{    
    glDeleteTextures(1, &videoFrameTexture);
    // in real world usage, we should delete textures whenever we can, to free up memory and GPU resources
    
    glDeleteTextures(1, &LUTTexture);
    
    glDeleteBuffers(2, vboIds);
}

#pragma mark -
#pragma mark OpenGL ES 2.0 rendering methods

- (void)drawFrame
{    
    glClear(GL_COLOR_BUFFER_BIT);
    
	// Use shader program.
	if (displayModeHasChanged) {
        switch (displayMode)
        {
            case PASSTHROUGH_VIDEO:
            {
                [glView setDisplayFramebuffer];
                glUseProgram(directDisplayProgram);
    
                // Update uniform values (MUST be after calling useProgram)
                glUniform1i(uniforms[UNIFORM_VIDEOFRAME], 0);   // bind to tex unit 0
                
            }; break;
            case SIMPLE_DITHERING:
            {
                [glView setDisplayFramebuffer];
                glUseProgram(ditherProgram);
                
                // Update uniform values (MUST be after calling useProgram)
                glUniform1i(uniforms[UNIFORM_VIDEOFRAME], 0);   // bind to tex unit 0
                glUniform1i(uniforms[UNIFORM_LUTFRAME], 1);     // bind to tex unit 1
                
                glUniform1fv(uniforms[UNIFORM_DITHERMATRIX],64, NL_ditherShader_ditherMatrix);
            }; break;
        }
        displayModeHasChanged = NO;
    }
    
    //glBindTexture(GL_TEXTURE_2D, videoFrameTexture);
    
    glDrawElements(GL_TRIANGLE_STRIP, 4, GL_UNSIGNED_SHORT, 0);
}

#pragma mark -
#pragma mark OpenGL ES 2.0 setup methods

- (BOOL)loadVertexShader:(NSString *)vertexShaderName fragmentShader:(NSString *)fragmentShaderName forProgram:(GLuint *)programPointer;
{
    GLuint vertexShader, fragShader;
	
    NSString *vertShaderPathname, *fragShaderPathname;
    
    // Create shader program.
    *programPointer = glCreateProgram();
    
    // Create and compile vertex shader.
    vertShaderPathname = [[NSBundle mainBundle] pathForResource:vertexShaderName ofType:@"vsh"];
    if (![self compileShader:&vertexShader type:GL_VERTEX_SHADER file:vertShaderPathname])
    {
        printf("Failed to compile vertex shader\n");
        return FALSE;
    }
    
    // Create and compile fragment shader.
    fragShaderPathname = [[NSBundle mainBundle] pathForResource:fragmentShaderName ofType:@"fsh"];
    if (![self compileShader:&fragShader type:GL_FRAGMENT_SHADER file:fragShaderPathname])
    {
        printf("Failed to compile fragment shader\n");
        return FALSE;
    }
    
    // Attach vertex shader to program.
    glAttachShader(*programPointer, vertexShader);
    
    // Attach fragment shader to program.
    glAttachShader(*programPointer, fragShader);
    
    // Bind attribute locations.
    // This needs to be done prior to linking.
    glBindAttribLocation(*programPointer, ATTRIB_VERTEX, "position");
    glBindAttribLocation(*programPointer, ATTRIB_TEXTUREPOSITON, "inputTextureCoordinate");
    
    // Link program.
    if (![self linkProgram:*programPointer])
    {
        printf("Failed to link program: %d\n", *programPointer);
        
        if (vertexShader)
        {
            glDeleteShader(vertexShader);
            vertexShader = 0;
        }
        if (fragShader)
        {
            glDeleteShader(fragShader);
            fragShader = 0;
        }
        if (*programPointer)
        {
            glDeleteProgram(*programPointer);
            *programPointer = 0;
        }
        
        return FALSE;
    }
    
    // Get uniform locations.
    uniforms[UNIFORM_VIDEOFRAME] = glGetUniformLocation(*programPointer, "videoFrame");
    uniforms[UNIFORM_LUTFRAME] = glGetUniformLocation(*programPointer, "lutFrame");
    
    // Release vertex and fragment shaders.
    if (vertexShader)
	{
        glDeleteShader(vertexShader);
	}
    if (fragShader)
	{
        glDeleteShader(fragShader);		
	}

    return TRUE;
}

- (BOOL)compileShader:(GLuint *)shader type:(GLenum)type file:(NSString *)file
{
    GLint status;
    const GLchar *source;
    
    source = (GLchar *)[[NSString stringWithContentsOfFile:file encoding:NSUTF8StringEncoding error:nil] UTF8String];
    if (!source)
    {
        printf("Failed to load vertex shader\n");
        return FALSE;
    }
    
    *shader = glCreateShader(type);
    glShaderSource(*shader, 1, &source, NULL);
    glCompileShader(*shader);
    
#if defined(DEBUG)
    GLint logLength;
    glGetShaderiv(*shader, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetShaderInfoLog(*shader, logLength, &logLength, log);
        printf("Shader compile log:\n%s\n", log);
        free(log);
    }
#endif
    
    glGetShaderiv(*shader, GL_COMPILE_STATUS, &status);
    if (status == 0)
    {
        glDeleteShader(*shader);
        return FALSE;
    }
    
    return TRUE;
}

- (BOOL)linkProgram:(GLuint)prog
{
    GLint status;
    
    glLinkProgram(prog);
    
#if defined(DEBUG)
    GLint logLength;
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        printf("Program link log:\n%s\n", log);
        free(log);
    }
#endif
    
    glGetProgramiv(prog, GL_LINK_STATUS, &status);
    if (status == 0)
        return FALSE;
    
    return TRUE;
}

- (BOOL)validateProgram:(GLuint)prog
{
    GLint logLength, status;
    
    glValidateProgram(prog);
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        printf("Program validate log:\n%s\n", log);
        free(log);
    }
    
    glGetProgramiv(prog, GL_VALIDATE_STATUS, &status);
    if (status == 0)
        return FALSE;
    
    return TRUE;
}

#pragma mark -
#pragma mark Display mode switching

- (void)handleSwitchOfDisplayMode:(id)sender;
{
    displayModeHasChanged = YES;
	displayMode = [sender selectedSegmentIndex];
}

#pragma mark -
#pragma mark Texture caching

- (GLuint)createTextureUsingWidth:(GLuint)w Height:(GLuint)h
{
	int dataSize = w * h * 4;
	uint8_t* textureData = malloc(dataSize);
	if(textureData == NULL)
		return 0;
	memset(textureData, 128, dataSize);
	
	GLuint handle;
    
    // Create a new texture from the camera frame data, display that using the shaders
    glGenTextures(1, &handle);
    glBindTexture(GL_TEXTURE_2D, handle);
    
    // setting parameters BEFORE we upload texture data
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
    // This is necessary for non-power-of-two textures
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
	
    // Using BGRA extension to pull in video frame data directly
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, w, h, 0, GL_BGRA, GL_UNSIGNED_BYTE, textureData);
    
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
	
    free(textureData);
	
	return handle;
}

#pragma mark -
#pragma mark Attribute and uniform initialization methods

- (void)initGLConstantValues;
{    
    typedef struct 
    {
        GLfloat x,  y,  // Vertex
                s0, t0; // Texcoord0
        
    } NL2dVertex;
    
    const NL2dVertex pVertices[4] = {   // arranged s.t. Vertices are on left, Texcoord0 is on right.
        -1.0, -1.0, 1.0, 1.0,
         1.0, -1.0, 1.0, 0.0,
        -1.0,  1.0, 0.0, 1.0,
         1.0,  1.0, 0.0, 0.0
    };
    
    const GLushort NL2dVertexIndices[4] = {
         0,1,2,3
    };
    
    glActiveTexture(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);

    // VBOs... refer to http://www.opengl.org/wiki/Vertex_Buffer_Object for help in managing and setting these up
    
    // vboIds[0] – used to store vertex position and texture coordinates
    // vboIds[1] – used to store element indices
    
    glGenBuffers(2, vboIds);
    
    glBindBuffer(GL_ARRAY_BUFFER, vboIds[0]);
    glBufferData(GL_ARRAY_BUFFER, (4*sizeof(GLfloat))*4, NULL, GL_STATIC_DRAW);
    glBufferSubData(GL_ARRAY_BUFFER, 0, (4*sizeof(GLfloat))*4, pVertices);
    
    glEnableVertexAttribArray(ATTRIB_VERTEX);
    glVertexAttribPointer(ATTRIB_VERTEX, 2, GL_FLOAT, GL_FALSE, 4*sizeof(GLfloat), (char*)NULL + 0);
    
    glEnableVertexAttribArray(ATTRIB_TEXTUREPOSITON);
    glVertexAttribPointer(ATTRIB_TEXTUREPOSITON, 2, GL_FLOAT, GL_FALSE, 4*sizeof(GLfloat), (char*)NULL + 2*(sizeof(GLfloat)));

    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, vboIds[1]);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4*sizeof(GLushort), NL2dVertexIndices, GL_STATIC_DRAW);
    
    glBindTexture(GL_TEXTURE_2D, videoFrameTexture);
}

- (UIImage*)loadLUTImageWithFilename:(NSString*)filename;
{
    NSArray* paths = NSSearchPathForDirectoriesInDomains(NSLibraryDirectory, NSUserDomainMask, YES);
    NSString* libraryDirectory = [paths objectAtIndex:0];
    
    NSString* path = [libraryDirectory stringByAppendingPathComponent:filename];
    
    if (path == nil) {
        NSLog(@"ERROR: Could not find image with filename %@",filename);
        return nil;
    }
    
    //NSLog(@"Path was found at %@",path);
    
    UIImage* image = [UIImage imageWithContentsOfFile:path];
    return image;
}

- (uint8_t*)getLUTAsBitmapFromUIImage:(UIImage*)LUTImage;
{
    // malloc a reasonably sized amount of memory on the heap to temporarily hold the texture
    uint8_t* LUTBitmap = malloc(sizeof(uint8_t) * (1024*32) * 4);
    
    // use CGBitmapContextCreate to turn it into a raw bitmap that we can feed directly to OpenGL
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
    CGContextRef context=CGBitmapContextCreate(LUTBitmap, 1024, 32, 8, 1024 * 4, colorSpace, kCGImageAlphaNoneSkipLast );
    
    if (context == nil) {
        CGColorSpaceRelease(colorSpace);

        printf("ERROR: Failed to load texture into context\n");
        return nil;
    }
    CGContextDrawImage(context, CGRectMake(0,0, 1024, 32), [LUTImage CGImage]);
    
    // cleanup
    CGContextRelease(context);
    CGColorSpaceRelease(colorSpace);
    
    return LUTBitmap;
}

- (void)loadLUTTexture;
{
    // Import LUTOutput.png - use Cocoa to load it as an autoreleased UIImage.    
    UIImage* LUTImage = [self loadLUTImageWithFilename:@"LUTOutput.png"];
    
    // Transform it to a Bitmap Context.
    uint8_t* LUTBitmap = [self getLUTAsBitmapFromUIImage:LUTImage];
    
    // Create a new Texture, and retain its handle as LUTTexture
    glActiveTexture(GL_TEXTURE1);
    // perform glEnableVertexAttribArray here?
    glEnable(GL_TEXTURE_2D);
    
    // Create a new texture from the camera frame data, display that using the shaders
    glGenTextures(1, &LUTTexture);
    glBindTexture(GL_TEXTURE_2D, LUTTexture);
    
    // setting parameters BEFORE we upload texture data
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	
    // This is necessary for non-power-of-two textures
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    
    // Using RGBA extension to pull in video frame data directly
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, 1024, 32, 0, GL_RGBA, GL_UNSIGNED_BYTE, LUTBitmap);
        
    // cleanup
    free(LUTBitmap);
        
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
}

#pragma mark -
#pragma mark NLGLCameraDelegate methods

- (void)cameraHasConnected;
{
	//printf("Connected to camera");
    
    // We don't want to set bounds for an AVFoundation preview layer; an alternative is already defined by our OpenGL code
    
    // Apart from a debug-style confirmation, we'll leave this callback be.
}

- (void)processNewCameraFrame:(CMSampleBufferRef)sampleBuffer;
{
    //pboIndex = HD_Mod3(pboIndex+1);
    //pboNextIndex = HD_Mod3(pboIndex+1);
    
    if(videoFrameTexture == 0)
	{
        videoFrameTexture = [self createTextureUsingWidth:640 Height:480];
        
        glBindTexture(GL_TEXTURE_2D, videoFrameTexture);
	}

    //glActiveTexture(GL_TEXTURE0);
    //glBindTexture(GL_TEXTURE_2D, videoFrameTexture);

    CVImageBufferRef cameraFrame = CMSampleBufferGetImageBuffer(sampleBuffer);  // returns an autoreleased buffer
    
    CVPixelBufferLockBaseAddress(cameraFrame, 0);

    glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 640, 480, GL_BGRA, GL_UNSIGNED_BYTE, CVPixelBufferGetBaseAddress(cameraFrame));
	
    [self drawFrame];
    
    if (videoRecorder.isRecording) {
        [videoRecorder writeSampleAtTime:CMSampleBufferGetOutputPresentationTimeStamp(sampleBuffer)];
    }
    
    CVPixelBufferUnlockBaseAddress(cameraFrame, 0);

    [glView presentFramebuffer];
            
    CFRelease(sampleBuffer);
}


#pragma mark -
#pragma mark Touch handling

- (void)touchesBegan:(NSSet *)touches withEvent:(UIEvent *)event
{
	//currentTouchPoint = [[touches anyObject] locationInView:self.view];
}

- (void)touchesMoved:(NSSet *)touches withEvent:(UIEvent *)event;
{
}

- (void)touchesEnded:(NSSet *)touches withEvent:(UIEvent *)event 
{   
    /*
    if (!videoRecorder.isRecording) {
        [videoRecorder startRecording];
    }
    else {
        [videoRecorder stopRecording];
    }
     */
    [camera openGLViewScreenshot:glView];
}

- (void)touchesCancelled:(NSSet *)touches withEvent:(UIEvent *)event 
{
}

#pragma mark -
#pragma mark Accessors

@synthesize glView;

@end
