//
// This file is part of AlphaCam II.
//
// AlphaCam II is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AlphaCam II is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AlphaCam II.  If not, see <http://www.gnu.org/licenses/>.
//
//
//  NLGLView.m
//  AlphaCam II
//
//  Created by Nicholas Levin on 5/6/11.
//  Copyright 2011 Studio Andromeda. All rights reserved.
//

#import "NLGLView.h"
#import <OpenGLES/EAGLDrawable.h>
#import <QuartzCore/QuartzCore.h>
#import "ESRenderer.h"

@implementation NLGLView

@synthesize delegate, backingWidth, backingHeight;

#pragma mark -
#pragma mark Initialization and teardown

// Override the class method to return the OpenGL layer, as opposed to the normal CALayer
+ (Class) layerClass 
{
	return [CAEAGLLayer class];
}


- (id)initWithFrame:(CGRect)frame 
{
    if ((self = [super initWithFrame:frame])) 
	{
		// Do OpenGL Core Animation layer setup
		CAEAGLLayer *eaglLayer = (CAEAGLLayer *)self.layer;
		
		// Set scaling to account for Retina display	
		if ([self respondsToSelector:@selector(setContentScaleFactor:)])
		{
            if ([[UIScreen mainScreen] scale] == 2.0)
                self.contentScaleFactor = 1.2;          // reduced to keep shader performance as good as it can be (1.35 works too, but w/ hit to fps)
                                                        //
                                                        // interestingly, this has the same width as the original Macintosh. Trivia/selling point worthy?
		}
		
		eaglLayer.opaque = YES;
		eaglLayer.drawableProperties = [NSDictionary dictionaryWithObjectsAndKeys:[NSNumber numberWithBool:NO], kEAGLDrawablePropertyRetainedBacking, kEAGLColorFormatRGBA8, kEAGLDrawablePropertyColorFormat, nil];		
		context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];
		
		if (!context || ![EAGLContext setCurrentContext:context] || ![self createFramebuffers]) 
		{
			return nil;
		}
    }
    return self;
}



- (void)drawView:(id)sender
{
    [delegate drawFrame];
}

#pragma mark -
#pragma mark OpenGL drawing

- (BOOL)createFramebuffers
{	
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_DEPTH_TEST);
    glDisable(GL_STENCIL_TEST);

    glEnable(GL_CULL_FACE);
    glCullFace(GL_FRONT_AND_BACK);
    
	// Onscreen framebuffer object
	glGenFramebuffers(1, &viewFramebuffer);
	glBindFramebuffer(GL_FRAMEBUFFER, viewFramebuffer);
	
	glGenRenderbuffers(1, &viewRenderbuffer);
	glBindRenderbuffer(GL_RENDERBUFFER, viewRenderbuffer);
	
	[context renderbufferStorage:GL_RENDERBUFFER fromDrawable:(CAEAGLLayer*)self.layer];
	
	glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_WIDTH, &backingWidth);
	glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_HEIGHT, &backingHeight);
	printf("Backing width: %d, height: %d\n", backingWidth, backingHeight);
	
	glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, viewRenderbuffer);
	
	if(glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE) 
	{
		printf("Failure with framebuffer generation\n");
		return NO;
	}
		
	GLenum status = glCheckFramebufferStatus(GL_FRAMEBUFFER);
    if (status != GL_FRAMEBUFFER_COMPLETE) 
	{
		printf("Incomplete FBO: %d\n", status);
        exit(1);
    }
	
	return YES;
}

- (void)destroyFramebuffer;
{	
	if (viewFramebuffer)
	{
		glDeleteFramebuffers(1, &viewFramebuffer);
		viewFramebuffer = 0;
	}
	
	if (viewRenderbuffer)
	{
		glDeleteRenderbuffers(1, &viewRenderbuffer);
		viewRenderbuffer = 0;
	}
}

- (void)setDisplayFramebuffer;
{
    if (context)
    {
        //[EAGLContext setCurrentContext:context];
        // only one context is ever used, so don't worry about switching between them
        
        if (!viewFramebuffer)
		{
            [self createFramebuffers];
		}
        
        //glBindFramebuffer(GL_FRAMEBUFFER, viewFramebuffer);   
        // our app already binds the only one we use when it calls -createFramebuffers
        
        //glViewport(0, 0, backingWidth, backingHeight);        
        // already defined in the View Controller's init
    }
}

- (BOOL)presentFramebuffer;
{
    BOOL success = FALSE;
    
    if (context)
    {
        //[EAGLContext setCurrentContext:context];
        // only one context is ever used, so don't worry about switching between them
        
        //glBindRenderbuffer(GL_RENDERBUFFER, viewRenderbuffer);
        // as with the frame buffer, the app binds the only render buffer we use when it calls -createFramebuffers
        
        success = [context presentRenderbuffer:GL_RENDERBUFFER];
    }
    
    return success;
}

#pragma mark -
#pragma mark Accessors

@synthesize positionRenderTexture;

@end
