//
// This file is part of AlphaCam II.
//
// AlphaCam II is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AlphaCam II is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AlphaCam II.  If not, see <http://www.gnu.org/licenses/>.
//
//
//  NLKDTree.h
//  AlphaCam II
//
//  Created by Nicholas Levin on 4/11/11.
//  Copyright 2011 Studio Andromeda. All rights reserved.
//

#import "NLImagePaletteStructures.h"

#define FSBALLOCATOR_USE_THREAD_SAFE_LOCKING_GCC_WITH_SCHED      // only if we require thread safety
#import "FSBAllocator.hh"

/* unbalanced kd-tree implementation translated to C++ 
 * and modified for integer precision (instead of int)
 * from java implementation in VideoMosaic
 * at http://www.intelegance.net/video/videomosaic.shtml
 *
 * refer to http://bisqwit.iki.fi/story/howto/dither/jy/
 * for more info on usage and hints for faster dithering
 */


template<typename V>
class KDTree
{
public:
    struct KDPoint
    {
        int coord[3];
        
        KDPoint() { }
        
        KDPoint(int a,int b,int c)
        {
            coord[0] = a; coord[1] = b; coord[2] = c;
        }
        KDPoint(int a,int b,int c,int d)
        {
            coord[0] = a; coord[1] = b; coord[2] = c; coord[3] = d;
        }
        
        KDPoint(int v[3])
        {
            for(unsigned n=0; n<3; ++n)
                coord[n] = v[n];
        }
        
        bool operator==(const KDPoint& b) const
        {
            for(unsigned n=0; n<3; ++n)
                if(coord[n] != b.coord[n]) return false;
            return true;
        }
        int sqrdist(const KDPoint& b) const
        {
            int result = 0;
            for(unsigned n=0; n<3; ++n)
            { int diff = coord[n] - b.coord[n];
                result += diff*diff; }
            return result;
        }
    };
private:
    struct KDRect
    {
        KDPoint min, max;
        
        KDPoint bound(const KDPoint& t) const
        {
            KDPoint p;
            for(unsigned i=0; i<3; ++i)
                if(t.coord[i] <= min.coord[i])
                    p.coord[i] = min.coord[i];
                else if(t.coord[i] >= max.coord[i])
                    p.coord[i] = max.coord[i];
                else
                    p.coord[i] = t.coord[i];
            return p;
        }
        void MakeInfinite()
        {
            for(unsigned i=0; i<3; ++i)
            {
                min.coord[i] = -1e99;
                max.coord[i] =  1e99;
            }
        }
    };
    
    struct KDNode;
    
    static KDNode*  NewNode(const KDNode&);
    static KDNode*  NewNode(const KDPoint& k, const V& v);
    static void DisposeNode(KDNode*);
    
    struct KDNode
    {
        KDPoint k;
        V       v;
        KDNode  *left, *right;
        
    public:
        KDNode() : k(),v(),left(0),right(0) { }
        KDNode(const KDPoint& kk, const V& vv)
        : k(kk), v(vv), left(0), right(0) { }
        
        virtual ~KDNode() { DisposeNode(left); DisposeNode(right); }
        
        static KDNode* ins( const KDPoint& key, const V& val,
                           KDNode*& t,
                           unsigned int lev)
        {
            if(!t)
                return (t = NewNode(key, val));
            else if(key == t->k)
                return 0; /* key duplicate */
            else if(key.coord[lev] > t->k.coord[lev])
                return ins(key, val, t->right, HD_Mod3(lev+1));
            else
                return ins(key, val, t->left,  HD_Mod3(lev+1));
        }
        
        struct Nearest
        {
            const KDNode* kd;
            int        dist_sqd;
        };
        // Method Nearest Neighbor from Andrew Moore's thesis. Numbered
        // comments are direct quotes from there. Step "SDL" is added to
        // make the algorithm work correctly.
        static void nnbr(const KDNode* kd, const KDPoint& target,
                         KDRect& hr, // in-param and temporary; not an out-param.
                         unsigned int lev,
                         Nearest& nearest)
        {
            // 1. if kd is empty then set dist-sqd to infinity and exit.
            if (!kd) return;
            
            // 2. s := split field of kd
            //int s = lev % K;
            
            // assuming K == 3,
            unsigned int s = HD_Mod3(lev);
            
            // 3. pivot := dom-elt field of kd
            const KDPoint& pivot = kd->k;
            int pivot_to_target = pivot.sqrdist(target);
            
            // 4. Cut hr into to sub-hyperrectangles left-hr and right-hr.
            //    The cut plane is through pivot and perpendicular to the s
            //    dimension.
            KDRect& left_hr = hr; // optimize by not cloning
            KDRect right_hr = hr;
            left_hr.max.coord[s]  = pivot.coord[s];
            right_hr.min.coord[s] = pivot.coord[s];
            
            // 5. target-in-left := target_s <= pivot_s
            bool target_in_left = target.coord[s] < pivot.coord[s];
            
            const KDNode* nearer_kd;
            const KDNode* further_kd;
            KDRect nearer_hr;
            KDRect further_hr;
            
            // 6. if target-in-left then nearer is left, further is right
            if (target_in_left) {
                nearer_kd = kd->left;
                nearer_hr = left_hr;
                further_kd = kd->right;
                further_hr = right_hr;
            }
            // 7. if not target-in-left then nearer is right, further is left
            else {
                nearer_kd = kd->right;
                nearer_hr = right_hr;
                further_kd = kd->left;
                further_hr = left_hr;
            }
            
            // 8. Recursively call Nearest Neighbor with parameters
            //    (nearer-kd, target, nearer-hr, max-dist-sqd), storing the
            //    results in nearest and dist-sqd
            nnbr(nearer_kd, target, nearer_hr, ++lev, nearest);
            
            // 10. A nearer point could only lie in further-kd if there were some
            //     part of further-hr within distance sqrt(max-dist-sqd) of
            //     target.  If this is the case then
            const KDPoint closest = further_hr.bound(target);
            if (closest.sqrdist(target) < nearest.dist_sqd)
            {
                // 10.1 if (pivot-target)^2 < dist-sqd then
                if (pivot_to_target < nearest.dist_sqd)
                {
                    // 10.1.1 nearest := (pivot, range-elt field of kd)
                    nearest.kd = kd;
                    // 10.1.2 dist-sqd = (pivot-target)^2
                    nearest.dist_sqd = pivot_to_target;
                }
                
                // 10.2 Recursively call Nearest Neighbor with parameters
                //      (further-kd, target, further-hr, max-dist_sqd)
                nnbr(further_kd, target, further_hr, ++lev, nearest);
            }
            // SDL: otherwise, current point is nearest
            else if (pivot_to_target < nearest.dist_sqd)
            {
                nearest.kd       = kd;
                nearest.dist_sqd = pivot_to_target;
            }
        }
    private:
        void operator=(const KDNode&);
    public:
        KDNode(const KDNode& b)
        : k(b.k), v(b.v),
        left( b.left ? NewNode(*b.left) : 0),
        right( b.right ? NewNode(*b.right) : 0 ) { }
    };
private:
    KDNode* m_root;
    size_t  m_count;
public:
    KDTree() : m_root(0), m_count(0) { }
    virtual ~KDTree() { DisposeNode(m_root); }
    
    void insert(const KDPoint& key, const V& val)
    {
        if(KDNode::ins(key, val, m_root, 0))
            m_count += 1;
    }
    
    const V* nearest(const KDPoint& key) const
    {
        KDRect hr;
        hr.MakeInfinite();
        
        typename KDNode::Nearest nn;
        nn.kd       = 0;
        nn.dist_sqd = 1e99;
        KDNode::nnbr(m_root, key, hr, 0, nn);
        if(nn.kd)
            return & nn.kd->v;
        return 0;
    }
    
    const std::pair<V,int>
    nearest_info(const KDPoint& key) const
    {
        KDRect hr;
        hr.MakeInfinite();
        
        typename KDNode::Nearest nn;
        nn.kd       = 0;
        nn.dist_sqd = 1e99;
        KDNode::nnbr(m_root, key, hr, 0, nn);
        if(nn.kd)
            return std::pair<V,int> ( nn.kd->v, nn.dist_sqd);
        return std::pair<V,int> ( V(), 1e99 );
    }
    
    size_t size() const { return m_count; }
public:
    KDTree& operator=(const KDTree&b)
    {
        if(this != &b)
        {
            if(m_root) DisposeNode(m_root);
            m_root = b.m_root ? NewNode(*b.m_root) : 0;
            m_count = b.m_count;
        }
        return *this;
    }
    KDTree(const KDTree& b)
    : m_root( b.m_root ? NewNode(*b.m_root) : 0 ),
    m_count( b.m_count ) { }
};

template<typename V>
typename KDTree<V>::KDNode* KDTree<V>::NewNode(const KDNode& b)
{
    KDNode* result = FSBAllocator<KDNode>().allocate(1);
    new (result) KDNode(b);
    return result;
}

template<typename V>
typename KDTree<V>::KDNode* KDTree<V>::NewNode(const KDPoint& k, const V& v)
{
    KDNode* result = FSBAllocator<KDNode>().allocate(1);
    new (result) KDNode(k,v);
    return result;
}

template<typename V>
void KDTree<V>::DisposeNode(KDNode* p)
{
    if(!p) return;
    p->~KDNode();
    FSBAllocator<KDNode>().deallocate(p, 1);
}
