//
// This file is part of AlphaCam II.
//
// AlphaCam II is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// AlphaCam II is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with AlphaCam II.  If not, see <http://www.gnu.org/licenses/>.
//
//
//  NLNearestColorLUTFactory.cpp
//  AlphaCam II
//
//  Created by Nicholas Levin on 5/22/11.
//  Copyright 2011 Studio Andromeda. All rights reserved.
//

#include "NLNearestColorLUTFactory.hh"

NLNearestColorLUTFactory::NLNearestColorLUTFactory(RGBPalette &inputPalette, bool useKDTree) {
    
    paletteTree = NULL;
    
    setInputPalette(inputPalette);
    activateKDTree(useKDTree);
}

NLNearestColorLUTFactory::~NLNearestColorLUTFactory() {
    
    if (paletteTree) {
        delete paletteTree;
    }
    
    free(imagePalette.table);
}

void NLNearestColorLUTFactory::setInputPalette(RGBPalette &inputPalette) {
    
    imagePalette.size = inputPalette.size;
    imagePalette.table = inputPalette.table;
    
    if (paletteTree) {
        for (unsigned i = 0; i < imagePalette.size; ++i) {
            paletteTree->insert( KDTree<unsigned char>::KDPoint( imagePalette.table[i].channel[0], imagePalette.table[i].channel[1], imagePalette.table[i].channel[2]), i );
        }
    }
}

void NLNearestColorLUTFactory::activateKDTree(bool useTree) {
    
    if (useTree) {
        if (!paletteTree) {
            paletteTree = new KDTree<unsigned char>();
        }
    }
    else {
        if (paletteTree) {
            delete paletteTree;
            paletteTree = NULL;
        }
    }
}

// compute for all RGB values, rather than as seperate channels (as we do when distributing the error)
const unsigned char NLNearestColorLUTFactory::findNearestColor(const uint8_t &red, const uint8_t &green, const uint8_t &blue) {
    
    if (paletteTree)
        return *paletteTree->nearest( KDTree<unsigned char>::KDPoint( red, green, blue ) );
    else {
        uint16_t bestIndex = 0;
        uint16_t approxDistance, minApproxDistance = 765;
        
        // more naive algorithm used here, because it's better suited to smaller palettes
        // (we can use the KD-Tree version for dithering palettes of larger size)
        
        for (unsigned i = 0, paletteLength = imagePalette.size; i < paletteLength; i++) {
            
            // preferred solution would be to use a LUT, but since our palettes tend to be < 257 colors, there's no real advantage to be had.
            // FLANN would be an acceptible solution if we were trying to match for 16-bit palettes, but that's not in the plans for version 1.0.
            
            approxDistance = abs(red - imagePalette.table[i].channel[0]) + abs(green - imagePalette.table[i].channel[1]) + abs(blue - imagePalette.table[i].channel[2]);
            
            if (approxDistance < minApproxDistance) {
                
                minApproxDistance = approxDistance;
                bestIndex = i;
            }
        }
        
        return bestIndex;
    }
}

void NLNearestColorLUTFactory::applyNearestColor(uint8_t &red,  uint8_t &green,  uint8_t &blue) {
	
	const unsigned char nearestColorIndex = findNearestColor(red,green,blue);
	
	red = imagePalette.table[nearestColorIndex].channel[0];
	green = imagePalette.table[nearestColorIndex].channel[1];
	blue = imagePalette.table[nearestColorIndex].channel[2];
}

void NLNearestColorLUTFactory::createLUT() {
    
    // consider the first input to be the red channel, second to be green, and third to be blue
    
    // initialize the outputArray with default values, and find the nearest matching color in the current palette
    int b, g, r;
    for (b = 0; b < 32; ++b) {
        for (g = 0; g < 32;  ++g) {
            for (r = 0; r < 32; ++r) {
                nearestColorLUT[r][g][b].channel[0] = (r << 3) + 4; // or * 8 + 4; (offset is to avoid biasing towards black)
                nearestColorLUT[r][g][b].channel[1] = (g << 3) + 4;
                nearestColorLUT[r][g][b].channel[2] = (b << 3) + 4;
                
                applyNearestColor(nearestColorLUT[r][g][b].channel[0], nearestColorLUT[r][g][b].channel[1], nearestColorLUT[r][g][b].channel[2]);
            }
        }
    }
    
    createdLUT = true;
}

void NLNearestColorLUTFactory::writeLUTToPNG(const char* filename) {
    
    if (createdLUT) {
        // transform our three dimensional array into a newly malloced, one dimensional interpolated array (either RGB or RGBA), assuming x = (r + (32 * g)) and y = b.
        uint8_t* LUTBitmap = static_cast<uint8_t *>(malloc(sizeof(uint8_t) * (32*32*32) * 4));
        
        // now process each bit into bitmap form!
        int b, g, r;
        for (b = 0; b < 32; ++b) {
            for (g = 0; g < 32;  ++g) {
                for (r = 0; r < 32; ++r) {
                    
                    // this is r (x), with g (x with offset) and b (y), collectively multiplied by the stride of our channels (4) and added
                    //  with an offset that represents their placement (RGB, without A)
                    
                    LUTBitmap[((r+(32*g)+(1024*b))*4)+0] = nearestColorLUT[r][g][b].channel[0];
                    LUTBitmap[((r+(32*g)+(1024*b))*4)+1] = nearestColorLUT[r][g][b].channel[1];
                    LUTBitmap[((r+(32*g)+(1024*b))*4)+2] = nearestColorLUT[r][g][b].channel[2];
                }
            }
        }
        
        // create a brand new bitmap context using the new array    
        CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
        CGContextRef context=CGBitmapContextCreate(LUTBitmap, 1024, 32, 8, 1024 * 4, colorSpace, kCGImageAlphaNoneSkipLast | kCGBitmapByteOrderDefault);
        
        // transform the context into an image
        const CGImageRef image = CGBitmapContextCreateImage(context);
        CGContextRelease(context);
        
        // finally, save it to the Documents directory, as demonstrated by ResultViewController.mm
        NL_Save_CGImage_As_PNG(image,filename,NL_Path_To_Library_Directory());
        CGImageRelease(image);
        
        // cleanup
        CGColorSpaceRelease(colorSpace);
        free(LUTBitmap);
    }
    else {
        printf("ERROR: Method writeNearestColorLUTToPNG(...) was called before a nearestColorLUT could be created.\n");
        exit(1);
    }
}
